/*******************************************************************************
* SMW_SX1276M0 Simple Receive - ABP (v1.0)
* 
* Simple program to test the downlink procedure with ABP.
* This program uses the ESP32 to communicate with the LoRaWAN module.
* 
* Copyright 2020 RoboCore.
* Written by Francois (17/07/20).
* 
* 
* This file is part of the SMW_SX1276M0 library ("SMW_SX1276M0-lib").
* 
* "SMW_SX1276M0-lib" is free software: you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* "SMW_SX1276M0-lib" is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU Lesser General Public License for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with "SMW_SX1276M0-lib". If not, see <https://www.gnu.org/licenses/>
*******************************************************************************/

#if !defined(ARDUINO_ESP32_DEV) // ESP32
#error Use this example with the ESP32
#endif

// --------------------------------------------------
// Libraries

#include "RoboCore_SMW_SX1276M0.h"

// --------------------------------------------------
// Variables

#include <HardwareSerial.h>
HardwareSerial LoRaSerial(2);
#define RXD2 16
#define TXD2 17

SMW_SX1276M0 lorawan(LoRaSerial);

CommandResponse response;

const char devaddr[] = "00000000";
const char appskey[] = "00000000000000000000000000000000";
const char nwkskey[] = "00000000000000000000000000000000";

boolean connected = false;
const unsigned long pause_time = 300000; // [ms] (5 min)
unsigned long timeout;
int count = 0;

// --------------------------------------------------
// Prototypes

void event_handler(Event);

// --------------------------------------------------
// --------------------------------------------------

void setup() {
  // Start the UART for debugging
  Serial.begin(115200);
  Serial.println(F("--- SMW_SX1276M0 Downlink (ABP) ---"));
  
  // start the UART for the LoRaWAN Bee
  LoRaSerial.begin(115200, SERIAL_8N1, RXD2, TXD2);

  // set the event handler
  lorawan.event_listener = &event_handler;
  Serial.println(F("Handler set"));

  // read the Device EUI
  char deveui[16];
  response = lorawan.get_DevEUI(deveui);
  if(response == CommandResponse::OK){
    Serial.print(F("DevEUI: "));
    Serial.write((uint8_t *)deveui, 16);
    Serial.println();
  } else {
    Serial.println(F("Error getting the Device EUI"));
  }

  // set the Device Address
  response = lorawan.set_DevAddr(devaddr);
  if(response == CommandResponse::OK){
    Serial.print(F("Device Address set ("));
    Serial.write((uint8_t *)devaddr, 8);
    Serial.println(')');
  } else {
    Serial.println(F("Error setting the Device Address"));
  }

  // set the Application Session Key
  response = lorawan.set_AppSKey(appskey);
  if(response == CommandResponse::OK){
    Serial.print(F("Application Session Key set ("));
    Serial.write((uint8_t *)appskey, 32);
    Serial.println(')');
  } else {
    Serial.println(F("Error setting the Application Session Key"));
  }

  // set the Network Session Key
  response = lorawan.set_NwkSKey(nwkskey);
  if(response == CommandResponse::OK){
    Serial.print(F("Network Session Key set ("));
    Serial.write((uint8_t *)nwkskey, 32);
    Serial.println(')');
  } else {
    Serial.println(F("Error setting the Network Session Key"));
  }

  // set join mode to ABP
  response = lorawan.set_JoinMode(SMW_SX1276M0_JOIN_MODE_ABP);
  if(response == CommandResponse::OK){
    Serial.println(F("Mode set to ABP"));
  } else {
    Serial.println(F("Error setting the join mode"));
  }

  // join the network (not really necessary in ABP)
  Serial.println(F("Joining the network"));
  lorawan.join();
}

// --------------------------------------------------
// --------------------------------------------------

void loop() {
  // listen for incoming data from the module
  lorawan.listen();

  // send a message
  if(connected){
    if(timeout < millis()){
      // update the counter
      count++;
      if(count > 255){
        count = 0; // reset
      }
  
      // convert to HEX
      char data[] = { '0', '0', 0};
      data[0] += (count / 16);
      if(data[0] > '9'){
        data[0] += 7;
      }
      data[1] += (count % 16);
      if(data[1] > '9'){
        data[1] += 7;
      }
  
      // send the message
      Serial.print(F("Data: "));
      Serial.println(data);
      response = lorawan.sendX(1, data);
  
      // update the timeout
      timeout = millis() + pause_time;
    }
  } else {
    if(timeout < millis()){
      // show some activity
      Serial.println('.');
    
      // update the timeout
      timeout = millis() + 5000; // 5 s
    }
  }
}

// --------------------------------------------------
// --------------------------------------------------

// Handle the events of the module
//  @param (type) : the type of the event [Event]
void event_handler(Event type){
  // check if join event
  if(type == Event::JOINED){
    Serial.println(F("Joined"));
    connected = true; // set
  } 
  // check if text message received event
  else if(type == Event::RECEIVED){
    Serial.println(F("Text message received"));
    uint8_t port;
    Buffer buffer;
    response = lorawan.readT(port, buffer);
    if(response == CommandResponse::OK){
      Serial.print(F("Message: "));
      while(buffer.available()){
        Serial.write(buffer.read());
      }
      Serial.print(F(" on port "));
      Serial.println(port);
    }
  } 
  // check if hexadecimal message received event
  else if(type == Event::RECEIVED_X){
    Serial.println(F("HEX message received"));
    uint8_t port;
    Buffer buffer;
    response = lorawan.readX(port, buffer);
    if(response == CommandResponse::OK){
      Serial.print(F("Message: "));
      while(buffer.available()){
        Serial.write(buffer.read());
      }
      Serial.print(F(" on port "));
      Serial.println(port);
    }
  }
}

// --------------------------------------------------
